package com.kameleoon.kameleoon_client_flutter

import android.app.Activity
import android.content.Context
import androidx.annotation.NonNull
import com.kameleoon.*
import com.kameleoon.KameleoonException
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import org.json.JSONObject
import org.json.JSONArray


/** KameleoonClientFlutterPlugin */
class KameleoonClientFlutterPlugin: FlutterPlugin, MethodCallHandler, ActivityAware {
  /// The MethodChannel that will the communication between Flutter and native Android
  ///
  /// This local reference serves to register the plugin with the Flutter Engine and unregister it
  /// when the Flutter Engine is detached from the Activity
  private lateinit var channel : MethodChannel

  private lateinit var activity: Activity
  private lateinit var context: Context
  private val clients = mutableMapOf<String, KameleoonClient>()

  override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    channel = MethodChannel(flutterPluginBinding.binaryMessenger, "kameleoon_client_flutter")
    channel.setMethodCallHandler(this)
    context = flutterPluginBinding.applicationContext
  }

  override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
    val siteCode = call.argument<String>("siteCode")
    val visitorCode = call.argument<String>("visitorCode")

    if (call.method.equals("initializeKameleoonClient")) {
      initializeKameleoonClient(siteCode!!, context)
      result.success("")
    }
    else if (call.method.equals("triggerExperiment")) {
      val experimentID = call.argument<Int>("experimentID")
      try {
        val variationID = clients[siteCode]?.triggerExperiment(visitorCode, experimentID!!)
        result.success(variationID)
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("flush")) {
      try {
        clients[siteCode]?.flush(visitorCode)
        result.success("")
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("trackConversion")) {
      val goalID = call.argument<Int>("goalID")
      val revenue = call.argument<Double>("revenue")?.toFloat()
      try {
        print("BBB >>> $revenue")
        clients[siteCode]?.trackConversion(visitorCode, goalID!!, revenue)
        result.success("")
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("activateFeature")) {
      val featureKey = call.argument<String>("featureKey")
      val featureID = call.argument<Int>("featureID")
      try {
        val feature = if (featureID != null) {
          clients[siteCode]?.activateFeature(visitorCode, featureID) == true
        } else {
          clients[siteCode]?.activateFeature(visitorCode, featureKey) == true
        }
        result.success(feature)
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("obtainVisitorCode")) {
      val defaultVisitorCode = call.argument<String>("defaultVisitorCode")
      if (defaultVisitorCode != null) {
        result.success(clients[siteCode]?.obtainVisitorCode(defaultVisitorCode))
      } else {
        result.success(clients[siteCode]?.obtainVisitorCode(null))
      }
    }
    else if (call.method.equals("obtainVariationAssociatedData")) {
      val variationID = call.argument<Int>("variationID")
      try {
        val obj: JSONObject? = clients[siteCode]?.obtainVariationAssociatedData(variationID!!)
        result.success(obj?.toMap() ?: JSONObject())
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("obtainFeatureVariable")) {
      val featureKey = call.argument<String>("featureKey")
      val featureID = call.argument<Int>("featureID")
      val variableKey = call.argument<String>("variableKey")

      try {
        val featureVariable = if (featureID != null) {
          clients[siteCode]?.obtainFeatureVariable(featureID, variableKey!!)
        } else {
          clients[siteCode]?.obtainFeatureVariable(featureKey!!, variableKey!!)
        }
        when (featureVariable) {
          is JSONArray -> result.success(featureVariable.toList())
          is JSONObject -> result.success(featureVariable.toMap());
          else -> result.success(featureVariable)
        }
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("addData")) {
      val dataType = call.argument<String>("dataType")
      val data = when (dataType) {
        "CustomData" -> {
          val id = call.argument<Int>("id")
          val value = call.argument<String>("value")
          Data.CustomData(id!!, value)
        }
        "PageView" -> {
          val url = call.argument<String>("url")
          val title = call.argument<String>("title")
          val referrer = call.argument<Int>("referrer")
          Data.PageView(url, title, referrer)
        }
        "Interest" -> {
          val index = call.argument<Int>("index")
          Data.Interest(index!!)
        }
        "Conversion" -> {
          val goalId = call.argument<Int>("goalId")
          val revenue = call.argument<Float>("revenue")
          val negative = call.argument<Boolean>("negative")
          Data.Conversion(goalId!!, revenue, negative)
        }
        "Browser" -> {
          val browser = call.argument<Int>("browser")
          try {
            Data.Browser.values()[browser!!] ?: null
          } catch (e: Exception) {
            throw e
          }
        }
        else -> {
          print("Data type not compatible")
          null
        }
      }
      try {
        if (data != null) {
          clients[siteCode]?.addData(visitorCode, data)
        }
        result.success(true);
      } catch (e: Exception) {
        result.error(getExceptionCode(e), e.message, e.stackTrace.toString())
      }
    }
    else if (call.method.equals("runWhenReady")) {
      result.success(null)
      val timeout = call.argument<Int>("timeout")
      class RunWhenReady : KameleoonReadyCallback {
        override fun onReady() {
          activity.runOnUiThread(java.lang.Runnable {
            channel.invokeMethod("setReady", true)
          })
        }

        override fun onTimeout() {
          activity.runOnUiThread(java.lang.Runnable {
            channel.invokeMethod("setReady", false)
          })
        }
      }

      clients[siteCode]?.runWhenReady(RunWhenReady(), timeout)
    }
    else if (call.method.equals("isReady")) {
      result.success(clients[siteCode]?.isReady);
    }
    else {
      result.notImplemented()
    }
  }

  override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
    channel.setMethodCallHandler(null)
  }

  private fun initializeKameleoonClient(siteCode: String, context: Context) {
    val kameleoonClient = KameleoonClientFactory.create(siteCode, context)
    //kameleoonClient.updateConfiguration();
    clients[siteCode] = kameleoonClient
  }

  override fun onAttachedToActivity(binding: ActivityPluginBinding) {
    activity = binding.activity
  }

  override fun onDetachedFromActivityForConfigChanges() {
    TODO("Not yet implemented")
  }

  override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
    TODO("Not yet implemented")
  }

  override fun onDetachedFromActivity() {
//    TODO("Not yet implemented")
  }
}

private fun KameleoonClientFlutterPlugin.getExceptionCode(exception: Exception): String =
  when (exception) {
    is KameleoonException.ExperimentConfigurationNotFound -> { "ExperimentConfigurationNotFound" }
    is KameleoonException.FeatureConfigurationNotFound -> { "FeatureConfigurationNotFound" }
    is KameleoonException.SDKNotReady -> { "SDKNotReady" }
    is KameleoonException.NotTargeted -> { "NotTargeted" }
    is KameleoonException.NotActivated -> { "NotActivated" }
    is KameleoonException.VariationConfigurationNotFound -> { "VariationConfigurationNotFound" }
    is KameleoonException.VisitorCodeNotValid -> { "VisitorCodeNotValid" }
    else -> { exception.toString() }
  }

private fun JSONArray.toList(): List<Map<String, *>> = 0.until(this.length()).map {
  when (val value = this[it]) {
    is JSONObject -> value.toMap()
    else -> emptyMap<String, Nothing>()
  }
}

private fun JSONObject.toMap(): Map<String, *> = keys().asSequence().associateWith {
  when (val value = this[it]) {
    is JSONArray -> {
      val map = 0.until(value.length()).associate { Pair(it.toString(), value[it]) }
      JSONObject(map).toMap().values.toList()
    }
    is JSONObject -> value.toMap()
    JSONObject.NULL -> null
    else            -> value
  }
}

