import 'dart:async';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:integration_test/integration_test.dart';
import 'package:kameleoon_client_flutter/kameleoon_client.dart';
import 'package:kameleoon_client_flutter/kameleoon_client_factory.dart';
import 'package:kameleoon_client_flutter/kameleoon_data.dart';
import 'package:kameleoon_client_flutter/kameleoon_exception.dart';

import 'constants.dart';

String makeReason(String text) => "\n\n\n" + ERROR_REASON + "\n" + text + "\n \n \n";


void main() {
  IntegrationTestWidgetsFlutterBinding.ensureInitialized();

  late final KameleoonClient client;

  testAPI([Function? actionWhenReady]) async {
    final completer = new Completer();

    client.runWhenReady(() async {
      await actionWhenReady?.call();
      completer.complete(COMPLETE_VALID);
    }, () {
      completer.completeError(TIMEOUT_FAIL_ERROR);
    }, TIMEOUT_MILLISECONDS);

    await expectLater(completer.future, completion(equals(COMPLETE_VALID)), reason: makeReason("runWhenReady Timeout"));
  }
  
  setUpAll(() => client = KameleoonClientFactory.create(siteCode)!);

  testWidgets("runWhenReadyTimeout", (WidgetTester tester) async {
    await testAPI();
  });

  testWidgets("isReady", (WidgetTester tester) async {
    await testAPI(() {
      expect(client.isReady(), equals(true), reason: makeReason("client.isReady() != true"));
    });
  });

  testWidgets("obtainFeatureVariable", (WidgetTester tester) async {
    await testAPI(() async {
      const featureId = 136230;
      const boolTrue = "boolTrueKey", boolFalse = "boolFalseKey", number = "numberKey", numberNull = "numberNullKey", string = "stringKey", json = "jsonKey", jsonArray = "jsonArrayKey";

      if (kIsWeb) { // web (obtainFeatureVariable returns only String)
        final variableByIdTrue = await client.obtainFeatureVariable(featureId, boolTrue);
        final variableByIdFalse = await client.obtainFeatureVariable(featureId, boolFalse);
        final variableByIdInt = await client.obtainFeatureVariable(featureId, number);
        final variableByIdIntNull = await client.obtainFeatureVariable(featureId, numberNull);
        final variableByIdString = await client.obtainFeatureVariable(featureId, string);
        final variableByIdJSON = await client.obtainFeatureVariable(featureId, json) as String;
        final variableByIdArrayJSON = await client.obtainFeatureVariable(featureId, jsonArray) as String;

        expect(variableByIdTrue, equals("True"), reason: makeReason("variableByIdTrue != true"));
        expect(variableByIdFalse, equals("False"), reason: makeReason("variableByIdFalse != false"));
        expect(variableByIdInt, equals("100"), reason: makeReason("variableByIdInt != 100"));
        expect(variableByIdIntNull, equals("0"), reason: makeReason("variableByIdIntNull != 0"));
        expect(variableByIdString, equals("TestString"), reason: makeReason("variableByIdString != TestString"));
        expect(jsonDecode(variableByIdJSON)['test'], equals('json'), reason: makeReason("variableByIdJSON['test'] != 'json'"));
        expect((jsonDecode(variableByIdArrayJSON) as List).length > 1 ? ((jsonDecode(variableByIdArrayJSON) as List)[1] as Map)['key'] : "", equals('value-2'), 
          reason: makeReason("((jsonDecode(variableByIdArrayJSON) as List)[1] as Map)['key'] != 'value-2'"));
      } else { // iOS & Android (obtainFeatureVariable returns different types)
        final variableByIdTrue = await client.obtainFeatureVariable(featureId, boolTrue) as bool;
        final variableByIdFalse = await client.obtainFeatureVariable(featureId, boolFalse) as bool;
        final variableByIdInt = await client.obtainFeatureVariable(featureId, number) as int;
        final variableByIdIntNull = await client.obtainFeatureVariable(featureId, numberNull) as int;
        final variableByIdString = await client.obtainFeatureVariable(featureId, string) as String;
        final variableByIdJSON = await client.obtainFeatureVariable(featureId, json) as Map;
        final variableByIdArrayJSON = await client.obtainFeatureVariable(featureId, jsonArray) as List;

        expect(variableByIdTrue, equals(true), reason: makeReason("variableByIdTrue != true"));
        expect(variableByIdFalse, equals(false), reason: makeReason("variableByIdFalse != false"));
        expect(variableByIdInt, equals(100), reason: makeReason("variableByIdInt != 100"));
        expect(variableByIdIntNull, equals(0), reason: makeReason("variableByIdIntNull != 0"));
        expect(variableByIdString, equals("TestString"), reason: makeReason("variableByIdString != TestString"));
        expect(variableByIdJSON['test'], equals('json'), reason: makeReason("variableByIdJSON['test'] != 'json'"));
        expect(variableByIdArrayJSON.length > 1 ? (variableByIdArrayJSON[1] as Map)['key'] : "", equals('value-2'), reason: makeReason("(variableByIdArrayJSON[1] as Map<String, Object>)['key'] != 'value-2'"));
      }

      const mockedVariationId = "toto";
      const featureKey = "testff";
      const mockedId = 139626;
      final variable139626ById = await client.obtainFeatureVariable(mockedId, mockedVariationId) as String;
      final variable139626ByKey = await client.obtainFeatureVariable(featureKey, mockedVariationId) as String;

      expect(variable139626ById, equals(variable139626ByKey), reason: makeReason("variable139626ById != variable139626ByKey"));
      expect(variable139626ById, equals(mockedVariationId), reason: makeReason("variable139626ById != mockedVariationId"));

      const fearureIdNotFound = 139626000;
      const fearureKeyNotFound = "not_found_feaure_key";
      const variableKey = "toto";
      const variableKeyNotFound = "not_found_variable_key";
      const reasonDesc = "obtainFeatureVariable must throw FeatureConfigurationNotFound";

      await expectLater(client.obtainFeatureVariable(fearureIdNotFound, variableKey), throwsA(isInstanceOf<FeatureConfigurationNotFound>()), reason: makeReason(reasonDesc));
      await expectLater(client.obtainFeatureVariable(fearureKeyNotFound, variableKey), throwsA(isInstanceOf<FeatureConfigurationNotFound>()), reason: makeReason(reasonDesc));
      await expectLater(client.obtainFeatureVariable(mockedId, variableKeyNotFound), throwsA(isInstanceOf<FeatureConfigurationNotFound>()), reason: makeReason(reasonDesc));
    });
  });

  testWidgets("obtainVariationAssociatedData", (WidgetTester tester) async {
    await testAPI(() async {
      const mockedVariationId = 617899;

      final data = await client.obtainVariationAssociatedData(mockedVariationId);

      final toto = data["toto"] as bool;
      final titi = data["titi"] as String;
      final tutu = data["tutu"] as bool;

      expect(toto && titi == 'tata' && !tutu, equals(true), reason: makeReason("(toto && titi == 'tata' && !tutu) == false") );

      const variationIdNotFound = 617899000;
      const reasonDesc = "obtainVariationAssociatedData must throw VariationConfigurationNotFound";      
      await expectLater(client.obtainVariationAssociatedData(variationIdNotFound), throwsA(isInstanceOf<VariationConfigurationNotFound>()), reason: makeReason(reasonDesc));
    });
  });

  testWidgets("triggerExperiments", (WidgetTester tester) async {
    await testAPI(() async {
      const visitorCode_2 = "test";
      const experimentId_1 = 136200, experimentId_2 = 136204, experimentId_3 = 136226;
      const expectedValue_1 = [0, 617899], expectedValue_2 = [617909, 617909], expectedValue_3 = [617982, 617983];
      const experimentId_Targeting = 136215;
      const expectedValue_Targeting = 0;
      final customData = CustomData(0, "true");

      final id136200_1 = await client.triggerExperiment(visitorCode, experimentId_1);
      final id136200_2 = await client.triggerExperiment(visitorCode_2, experimentId_1);
      expect(listEquals([id136200_1, id136200_2], expectedValue_1), equals(true), reason: makeReason("id136200_1, id136200_2] != expectedValue_1"));

      final id136204_1 = await client.triggerExperiment(visitorCode, experimentId_2);
      final id136204_2 = await client.triggerExperiment(visitorCode_2, experimentId_2);
      expect(listEquals([id136204_1, id136204_2], expectedValue_2), equals(true), reason: makeReason("id136204_1, id136204_2] != expectedValue_2"));

      final id136226_1 = await client.triggerExperiment(visitorCode, experimentId_3);
      final id136226_2 = await client.triggerExperiment(visitorCode_2, experimentId_3);
      expect(listEquals([id136226_1, id136226_2], expectedValue_3), equals(true), reason: makeReason("id136226_1, id136226_2] != expectedValue_3"));

      await expectLater(client.triggerExperiment(visitorCode_2, experimentId_Targeting), throwsA(isInstanceOf<NotTargeted>()));
      await client.addData(visitorCode_2, customData);
      final id136215 = await client.triggerExperiment(visitorCode_2, experimentId_Targeting);
      expect(id136215, equals(expectedValue_Targeting), reason: makeReason("id136215 != expectedValue_Targeting"));
    });
  });

  testWidgets("activateFeatureNotTargeted", (WidgetTester tester) async {
    await testAPI(() async {
      const mockedFeatureId = 139626;
      await expectLater(client.activateFeature(visitorCode, mockedFeatureId), throwsA(isInstanceOf<NotTargeted>()), reason: makeReason("activateFeature must throw NotTargeted"));
    });
  });

  testWidgets("trackConversion", (WidgetTester tester) async {
    await testAPI(() async {
      const mockedGoalId = 238420;
      await client.trackConversion(visitorCode, mockedGoalId, 10);
      await client.trackConversion(visitorCode, mockedGoalId);
    });
  });

  testWidgets("activateFeatureAddData", (WidgetTester tester) async {
    await testAPI(() async {
      const mockedfeatureId = 136230;
      final customData = CustomData(2, "true");
      await client.addData(visitorCode, customData);
      expect(await client.activateFeature(visitorCode, mockedfeatureId), equals(true), reason: makeReason("activateFeature != null"));
    });
  });

  testWidgets("flush", (WidgetTester tester) async {
    await testAPI(() async {
      await client.addData(visitorCode, CustomData(10, "afd"));
      await client.addData(visitorCode, Browser(Browsers.OPERA));
      await client.addData(visitorCode, PageView("dfjksjdf", "title", 1231));
      await client.flush();
    });
  });

  testWidgets("visitorCode", (WidgetTester tester) async {
    await testAPI(() async {
      if (kIsWeb) { // only iOS & Android
        return;
      }
      const VISITOR_CODE_MAX_LENGTH = 255;
      final visitorCodeTheLengthLimit = "a" * VISITOR_CODE_MAX_LENGTH;
      final visitorCodeTooLong = "a" * (VISITOR_CODE_MAX_LENGTH + 1);
      const visitorCodeEmpty = "";

      const mockedGoalId = 238420;
      final customData = CustomData(2, "true");
      const experimentId = 136200;
      const featureId = 136230;
      const featureKey = "test-feature-variables";

      checkVisitorCodeNotValid(String exception, Function testFunc) async {
        try {
          await testFunc.call();
          if(exception.isNotEmpty) {
            expectSync(true, equals(false), reason: makeReason("section '$exception' must throw expection"));
          }
        } on VisitorCodeNotValid catch(e) {
          expectSync(e.toString(), contains(exception), reason: makeReason("section '$exception'\n${e.toString()} must contain $exception\nwrong throw VisitorCodeNotValid expection"));
        } catch (e) {
          expectSync(true, equals(false), reason: makeReason("section '$exception'\n${e.toString()} must contain $exception\nthrow non VisitorCodeNotValid expection"));
        }
      }

      final mapLengthError = {
        visitorCodeTheLengthLimit: "",
        visitorCodeTooLong: "max length",
        visitorCodeEmpty: "is empty"
      };

      mapLengthError.forEach((visitorCode, exception) {
        checkVisitorCodeNotValid(exception, () async => await client.trackConversion(visitorCode, mockedGoalId) );
        checkVisitorCodeNotValid(exception, () async => await client.trackConversion(visitorCode, mockedGoalId, 10) );
        checkVisitorCodeNotValid(exception, () async => await client.addData(visitorCode, customData) );
        checkVisitorCodeNotValid(exception, () async => await client.activateFeature(visitorCode, featureId) );
        checkVisitorCodeNotValid(exception, () async => await client.activateFeature(visitorCode, featureKey) );
        checkVisitorCodeNotValid(exception, () async => await client.flush(visitorCode) );
        checkVisitorCodeNotValid(exception, () async => await client.triggerExperiment(visitorCode, experimentId) );
      });
    });
  });
}
